(* $Id: wtool.ml 1168 2012-12-29 01:01:45Z gerd $ *)

(* TODO:
   - set_proxy_from_environment: this method should be fixed so that it
     also looks into ftp_proxy. Needs per-transport proxy settings. - DONE
   - export Http_fs.translate - DONE
   - fs # read_file, fs # write_file - DONE
   - add authentication
      - commit, save - DONE
   - enter passwords for win32 - DONE
   - enter passwords under godi_console UI - DONE (?)
   - manage certificates - DONE
   - Handling of the domain list for digest authentication
     (after fixing in Http_client)
 *)

open Wtool_types
open Printf

let string_of_exn e =
  match e with
    | Failure msg -> 
	msg
    | Unix.Unix_error(code,fn,path) ->
	if path = "" then
	  sprintf "%s (function: %s)" 
	    (Unix.error_message code)
	    fn
	else
	  sprintf "%s: %s (function: %s)" 
	    path
	    (Unix.error_message code)
	    fn
    | e ->
	"Exception " ^ Netexn.to_string e


module Password = struct
  type entry =
      { proto : string;   (* "http", "https", "ftp" *)
	domain : string;  (* This is an URL *)
	realm : string;   (* only http, https. For ftp this is empty *)
	user : string;
	password : string;
      }

  type cache =
      { mutable stock : entry list;
	mutable added : entry list;
	mutable modified : bool;
	mutable en_store : bool;
      }

  let pw_file () : Godi_file.filename =
    (* FIXME: Win32 *)
    let p_uid = Unix.getuid() in
    let home_dir =
      try Sys.getenv "HOME"
      with Not_found ->
	try (Unix.getpwuid p_uid).Unix.pw_dir
	with Not_found ->
	  failwith "Cannot determine home directory of current user" in
    let home_dir_file = Godi_file.filename home_dir in
    Godi_file.addname home_dir_file ".godi_pw"

  let load_entries ?(en_store=false) () =
    let fn = pw_file() in
    if Godi_file.file_exists fn then (
      let ini = new Godi_inifiles.inifile (Godi_file.nativename fn) in
      let stock =
	List.map
	  (fun sect ->
	     { proto = ini#getval sect "protocol";
	       domain = ini#getval sect "domain";
	       realm = ini#getval sect "realm";
	       user = ini#getval sect "user";
	       password = Netencoding.Base64.decode (ini#getval sect "password")
	     }
	  )
	  ini#sects in
      { stock = stock;
	added = [];
	modified = false;
	en_store = en_store;
      }
    )
    else 
      { stock = [];
	added = [];
	modified = false;
	en_store = en_store;
      }

  let uniq l =
    (* Later entries in l beat earlier ones! *)
    let ht = Hashtbl.create 11 in
    List.iter
      (fun e ->
	 Hashtbl.replace ht (e.proto,e.domain,e.realm,e.user) e
      )
      l;
    Hashtbl.fold
      (fun _ v acc -> v :: acc)
      ht
      []

  let save_entries cache =
    if cache.modified && cache.en_store then (
      prerr_endline "[Saving password]";
      let fn = pw_file() in
      Godi_file.write_file_lines ~privacy:true ~mode:`Binary fn [];
      let ini = new Godi_inifiles.inifile (Godi_file.nativename fn) in
      List.iter
	(fun e ->
	   let sect =
	     Digest.to_hex
	       (Digest.string
		  (String.concat "\t"
		     [ e.proto; e.domain; e.realm; e.user ]
		  )
	       ) in
	   ini # setval sect "protocol" e.proto;
	   ini # setval sect "domain" e.domain;
	   ini # setval sect "realm" e.realm;
	   ini # setval sect "user" e.user;
	   ini # setval sect "password" 
	     (Netencoding.Base64.encode e.password)
	)
	cache.stock;
      cache.modified <- false;
      ini#save()
    )

  let commit cache =
    let all = uniq(cache.stock @ List.rev cache.added) in
    cache.modified <- cache.modified || cache.added <> [];
    cache.stock <- all;
    cache.added <- []

  let query_password cache e =
    (* In e, everything is set except password. This does not save the cache! *)
    let p = sprintf "Password for %s at %s? " e.user e.domain in
    let pw =
      if Godi_concmd.have_console() then (
	let tdir = Godi_file.allocate_tmpdir() in
	let f = 
	  Godi_file.nativename
	    (Godi_file.addname
	       (Godi_file.tmpdir_path tdir)
	       "wtool-prompt") in
	try
	  print_string
	    (Godi_concmd.string_of_concmd
	       (`Password_prompt(p, f)));
	  flush stdout;
	  let pw = Godi_concmd.receive_data f in
	  Godi_file.release_tmpdir tdir;
	  pw
	with
	  | error ->
	      Godi_file.release_tmpdir tdir;
	      raise error
      )
      else (
	let have_unix_tty =
	  match Sys.os_type with
	    | "Win32" -> false
	    | _ -> Unix.isatty Unix.stdin in
	if have_unix_tty then
	  Netsys_posix.tty_read_password
	    p
	else (
	  prerr_string p;
	  flush stderr;
          Godi_exec.input_line()
	)
      ) in
    let e1 = { e with password = pw } in
    cache.added <- e1 :: cache.added;
    pw
      
  let norm_domain dom =
    dom (* FIXME *)

  let ftp_get_password cache domain user =
    let domain = norm_domain domain in
    try
      let e =
	List.find
	  (fun e ->
	     e.proto = "ftp" && e.domain = domain && e.user = user
	  )
	  cache.stock in
      e.password
    with
      | Not_found ->
	  let e =
	    { proto = "ftp";
	      domain = domain;
	      realm = "";
	      user = user;
	      password = "" 
	    } in
	  query_password cache e

  let http_key_handler cache proto user : Http_client.key_handler =
    ( object
	method inquire_key ~domain ~realms ~auth =
	  let domain = norm_domain (List.hd domain) in (* CHECK *)
	  let password, realm =
	    try
	      let e =
		List.find
		  (fun e ->
		     e.proto = proto && 
		      e.domain = domain && 
		      List.mem e.realm realms &&
		      e.user = user
		  )
		  cache.stock in
	      (e.password, e.realm)
	    with
	      | Not_found ->
		  let realm = List.hd realms in (* CHECK *)
		  let e =
		    { proto = proto;
		      domain = domain;
		      realm = realm;
		      user = user;
		      password = "" 
		    } in
		  (query_password cache e, realm) in
	  ( object
	      method user = user
	      method password = password
	      method realm = realm
	      method domain = [domain]  (* CHECK URIs *)
	    end  : Http_client.key
	  )

	method invalidate_key _ = ()
      end
    )

end


module Certificate = struct
  type entry =
      { domain : string;  (* This is a DNS name *)
        fingerprint : string;
      }

  type cache =
      { mutable file : Godi_file.filename;
        mutable stock : entry list;
	mutable added : entry list;
	mutable modified : bool;
      }

  let cert_file localbase : Godi_file.filename =
    let f = Godi_file.filename localbase in
    Godi_file.addname (Godi_file.addname f "etc") "certificates"

  let load_entries localbase =
    let fn = cert_file localbase in
    if Godi_file.file_exists fn then (
      let ini = new Godi_inifiles.inifile (Godi_file.nativename fn) in
      let stock =
	List.map
	  (fun sect ->
	     { domain = ini#getval sect "domain";
               fingerprint = ini#getval sect "fingerprint";
	     }
	  )
	  ini#sects in
      { file = fn;
        stock = stock;
	added = [];
	modified = false;
      }
    )
    else 
      { file = fn;
        stock = [];
	added = [];
	modified = false;
      }


  let save_entries cache =
    if cache.modified then (
      prerr_endline "[Saving certificates]";
      let fn = cache.file in
      Godi_file.write_file_lines ~privacy:true ~mode:`Binary fn [];
      let ini = new Godi_inifiles.inifile (Godi_file.nativename fn) in
      List.iter
	(fun e ->
	   let sect =
	     Digest.to_hex
	       (Digest.string
		  (String.concat "\t"
		     [ e.domain; e.fingerprint ]
		  )
	       ) in
	   ini # setval sect "domain" e.domain;
	   ini # setval sect "fingerprint" e.fingerprint;
	)
	cache.stock;
      cache.modified <- false;
      ini#save()
    )

  let commit cache =
    let all = cache.stock @ List.rev cache.added in
    cache.modified <- cache.modified || cache.added <> [];
    cache.stock <- all;
    cache.added <- []

  let verify cache domain ctx ssl_sock fd =
    let cert = Ssl.get_certificate ssl_sock in
    let fp = Ssl_exts.certificate_fingerprint cert in
    
    let ok =
      List.exists
        (fun entry ->
           entry.domain = domain && entry.fingerprint = fp
        )
        cache.stock in
 
    if not ok then (
      let msg =
        sprintf "Accept certificate %s for %s (yes/no)? " fp domain in
      let ack_s =
        if Godi_concmd.have_console() then (
	  let tdir = Godi_file.allocate_tmpdir() in
	  let f = 
	    Godi_file.nativename
	      (Godi_file.addname
	         (Godi_file.tmpdir_path tdir)
	         "wtool-prompt") in
	  try
	    print_string
	      (Godi_concmd.string_of_concmd
	         (`Prompt(msg, f)));
	    flush stdout;
	    let ack_s = Godi_concmd.receive_data f in
            Godi_file.release_tmpdir tdir;
            ack_s
	  with
	    | error ->
	        Godi_file.release_tmpdir tdir;
	        raise error
        )
        else (
	  prerr_string msg;
	  flush stderr;
          Godi_exec.input_line()
        )
      in
      let ack = 
        match ack_s with
          | "yes" | "YES" | "y" | "Y" -> true
          | _ -> false in
      if not ack then
        failwith "Certificate rejected by user";
      let e = { domain; fingerprint = fp } in
      cache.added <- e :: cache.added;
      ( try
          commit cache;
          save_entries cache
        with 
          | error ->
              eprintf "Warning: cannot save certificates\n%!";
              eprintf "Exception: %s\n%!" (Netexn.to_string error)
      )
    )

end



let tee out1 out2 : Netchannels.raw_out_channel =
  let rec output_string ch s p n =
    let k = ch#output s p n in
    if k < n then
      output_string ch s (p+k) (n-k) in

  let p = ref 0 in
  ( object(self)
      method output s pos len =
        output_string out1 s pos len;
        out1#flush();
        output_string out2 s pos len;
        out2#flush();
        p := !p + len;
        len
      method flush() =
        out1#flush();
        out2#flush()
      method close_out() =
        out1#close_out();
        out2#close_out()
      method pos_out = !p
    end
  )


let config_pipeline p =
(* Http_client.Debug.enable := true; *)
  let ftp_syn = Hashtbl.find Neturl.common_url_syntax "ftp" in
  let opts = p # get_options in
  let opts' =
    { opts with
        Http_client.schemes = opts.Http_client.schemes @
        [ "ftp", ftp_syn, Some 21, Http_client.proxy_only_cb_id ]
    } in
  p # set_options opts';
  p # set_transport_proxy_from_environment
    [ "http_proxy",  Http_client.http_cb_id;
      "https_proxy", Http_client.https_cb_id;
      "ftp_proxy",   Http_client.proxy_only_cb_id
    ]


let use_web_proxy url =
  (* determines whether to use a web proxy for [url] *)
  try
    let p = new Http_client.pipeline in
    config_pipeline p;
    p # proxy_type url <> None
  with
    | _ -> false


let exit_re =
  Str.regexp "^### EXIT \\([0-9]+\\)$"


let http_handler ~localbase 
                 ~timeout ~user ~en_store ~proto base_url : wtool_handler =
  let certcache = Certificate.load_entries localbase in
  let verify =
    let neturl = Neturl.parse_url base_url in
    let domain = Neturl.url_host neturl in
    Certificate.verify certcache domain in
  let ctx = Ssl.create_context Ssl.TLSv1 Ssl.Client_context in
  let tct = Https_client.https_transport_channel_type ~verify ctx in
  let pwcache = Password.load_entries ~en_store () in
  let fs =
    Http_fs.http_fs
      ~config_pipeline:(
	fun p ->
	  config_pipeline p;
	  let opts = p # get_options in
	  let opts' =
	    { opts with
		Http_client.connection_timeout = timeout;
	    } in
	  p # set_options opts';
	  ( match user with
	      | Some u when proto = "http" || proto = "https" ->
		  let kh = Password.http_key_handler pwcache proto u in
		  let auth_handler = new Http_client.unified_auth_handler kh in
		  p # add_auth_handler auth_handler;
	      | _ -> 
		  ()
	  );
	  p # configure_transport Http_client.https_cb_id tct
      )
      ~streaming:true
      base_url in

  let pipeline = fs#pipeline in

  ( object(self)
      method base_url = base_url

      method private have_webdav path =
	(* Check whether we have WebDAV: *)
	(proto = "http" || proto = "https") && (
	  let opts = new Http_client.options (fs#translate path) in
	  opts # set_reconnect_mode Http_client.Send_again;
	  pipeline # add opts;
	  pipeline # run();
	  match opts # status with
	    | `Successful ->
		Password.commit pwcache;
		( try ignore(opts#response_header#field "DAV"); true
		  with Not_found -> false
		)
	    | `Client_error ->
		(* do not accept a 401 here (wrong password) *)
		if opts # response_status_code = 401 then
		  failwith "Permission denied"
		else
		  false
	    | _ ->
		false
	)

      method list path =
	try
	  let have_webdav = self # have_webdav path in
	  let files =
	    if have_webdav then (
	      (* For WebDAV we have the precise PROPFIND method for listings *)
	      let webdav_fs =
		Webdav_netfs.webdav_netfs_layer fs in
	      webdav_fs # readdir [] path
	    )
	    else
	      (* No WebDAV: do an ordinary GET on the URL, and look for
		 hyperlinks, in the hope that the page is a directory index.
	       *)
	      fs # readdir [] path in
	  Password.commit pwcache;
	  Password.save_entries pwcache;
	  `Ok
	    (List.filter
	       (fun n -> n <> "" && n.[0] <> '.')
	       files
	    )
	with
	  | Http_client.Http_protocol (Http_client.Timeout _) ->
	      `Timeout
	  | Http_client.Http_protocol e ->
	      `Error (string_of_exn e)
	  | e ->
	      `Error (string_of_exn e)

      method get tdir paths =
	(* don't use fs, but the pipeline directly *)
	let n = ref 0 in
	let rlist =
	  List.map
	    (fun path ->
	       let r = ref (`Error "unserved") in
	       let url = fs#translate path in
	       let call = new Http_client.get url in
	       let local_filename =
		 Godi_file.addname
		   (Godi_file.tmpdir_path tdir)
		   ("wtool-" ^ string_of_int !n) in
	       let local_filename_n =
		 Godi_file.nativename local_filename in
	       incr n;
	       call # set_response_body_storage
		 (`File(fun () -> local_filename_n));
	       pipeline # add_with_callback
		 call
		 (fun rcall ->
		    match rcall#status with
		      | `Unserved -> assert false
		      | `Http_protocol_error (Http_client.Timeout _) ->
			  r := `Timeout
		      | `Http_protocol_error e ->
			  r := `Error(string_of_exn e)
		      | `Redirection | `Client_error | `Server_error ->
			  r := `Error("HTTP code " ^ 
					string_of_int
					(rcall#response_status_code))
		      | `Successful ->
			  r := `Ok local_filename;
		 );
	       r
	    )
	    paths in
	Unixqueue.run pipeline#event_system;
	let all_ok =
	  List.for_all
	    (fun r -> match !r with `Ok _ -> true | _ -> false)
	    rlist in
	if all_ok then (
	  Password.commit pwcache;
	  Password.save_entries pwcache;
	);
	List.map (fun rref -> !rref) rlist

      method put local_filename path =
	try
	  let have_webdav = self # have_webdav path in
	  if have_webdav then (
	    let webdav_fs =
	      Webdav_netfs.webdav_netfs_layer fs in
	    let local_obj =
	      ( object
		  method filename = Godi_file.nativename local_filename
		  method close() = ()
		end
	      ) in
	    webdav_fs # write_file [ `Create; `Truncate ] path local_obj;
	    Password.commit pwcache;
	    Password.save_entries pwcache;
	    `Ok ()
	  )
	  else
	    failwith "Cannot PUT files to a server lacking WebDAV support"
	with 
	  | Http_client.Timeout _ -> `Timeout
	  | e -> `Error(string_of_exn e)


      method api args =
	let call = new Http_client.post base_url args in

        let buf = Buffer.create 800 in
        let buf_ch = new Netchannels.output_buffer buf in
        let stdout_ch = new Netchannels.output_channel stdout in
        let sync_ch = tee buf_ch stdout_ch in
        let async_ch = new Uq_engines.pseudo_async_out_channel sync_ch in
        let dev = `Async_out(async_ch, pipeline#event_system) in
        let r = ref(`Error "unserved") in
	call # set_response_body_storage
	  (`Device(fun () -> 
                     if call#response_status_code = 200 then
                       dev
                     else
                       let null_async_ch =
                         new Uq_engines.pseudo_async_out_channel
                           (new Netchannels.output_null()) in
                       `Async_out(null_async_ch, pipeline#event_system)
                  ));
        pipeline # add_with_callback
	  call
	  (fun rcall ->
	     match rcall#status with
	       | `Unserved -> assert false
	       | `Http_protocol_error (Http_client.Timeout _) ->
		   r := `Timeout
	       | `Http_protocol_error e ->
		   r := `Error(string_of_exn e)
	       | `Redirection | `Client_error | `Server_error ->
		   r := `Error("HTTP code " ^ 
				 string_of_int
				 (rcall#response_status_code))
	       | `Successful ->
                   let s = Buffer.contents buf in
                   let lines =
                     Netchannels.lines_of_in_obj_channel
                       (new Netchannels.input_string s) in
                   let exit_code = ref 1 in
                   List.iter
                     (fun line ->
                        if Str.string_match exit_re line 0 then
                          exit_code := int_of_string (Str.matched_group 1 line)
                     )
                     lines;
		   r := `Ok (!exit_code);
	  );
	Unixqueue.run pipeline#event_system;
        let all_ok =
          match !r with
            | `Ok _ -> true
            | _ -> false in
	if all_ok then (
	  Password.commit pwcache;
	  Password.save_entries pwcache;
	);
        !r
    end
  )

let ftp_handler ~timeout ~user ~en_store base_neturl : wtool_handler =
  let base_neturl1 =
    Neturl.make_url
      ~scheme:"ftp"
      ?user
      ~host:(Neturl.url_host base_neturl)
      ~port:(try Neturl.url_port base_neturl with Not_found -> 21)
      ~path:(try Neturl.url_path base_neturl with Not_found -> [""] )
      (Neturl.url_syntax_of_url base_neturl) in
  let base_url =
    Neturl.string_of_url base_neturl1 in
  let pwcache = Password.load_entries ~en_store () in
  let get_password =
    match user with
      | None -> (fun _ -> "")
      | Some _ ->
	  let domain = 
	    Neturl.string_of_url
	      (Neturl.remove_from_url ~path:true base_neturl1) in
	  (fun u -> Password.ftp_get_password pwcache domain u) in

  let with_tmpdir_fs tmpdir f =
    let fs =
      Ftp_fs.ftp_fs
	~config_client:(fun c ->
			  c # configure_timeout timeout
		       )
	~tmp_directory:(Godi_file.nativename(Godi_file.tmpdir_path tmpdir))
	~tmp_prefix:"wtooltmp-"
	~get_password
	~keep_open:true
	base_url in
    try
      let r = f fs in
      fs # close();
      r
    with
      | error ->
	  fs # close();
	  raise error in

  let with_fs f =
    let tmpdir = Godi_file.allocate_tmpdir() in
    try
      let r = with_tmpdir_fs tmpdir f in
      Godi_file.release_tmpdir tmpdir;
      r
    with
      | error ->
	  Godi_file.release_tmpdir tmpdir;
	  raise error in

  ( object(self)
      method base_url = base_url

      method list path =
	try
	  with_fs 
	    (fun fs ->
	       let files = fs # readdir [] path in
	       Password.commit pwcache;
	       Password.save_entries pwcache;
	       `Ok (List.filter
		      (fun n -> n <> "" && n.[0] <> '.')
		      files
		   )
	    )
	with
	  | Ftp_client.FTP_timeout _ ->
	      `Timeout
	  | error ->
	      `Error(string_of_exn error)

      method get tmpdir paths =
	try
	  with_tmpdir_fs
	    tmpdir
	    (fun fs ->
	       List.map
		 (fun path ->
		    try
		      let local =
			fs # read_file [ `Binary ] path in
		      Password.commit pwcache;
		      Password.save_entries pwcache;
		      `Ok (Godi_file.filename (local#filename))
			(* We don't call [close] ! *)
		    with
		      | Ftp_client.FTP_timeout _ ->
			  `Timeout
		      | error ->
			  `Error(string_of_exn error)
		 )
		 paths
	    )
	with
	  | Ftp_client.FTP_timeout _ ->
	      List.map (fun _ -> `Timeout) paths
	  | error ->
	      let code =
		`Error(string_of_exn error) in
	      List.map (fun _ -> code) paths


      method put local_filename path =
	try
	  with_fs 
	    (fun fs ->
	       let local =
		 ( object
		     method filename = Godi_file.nativename local_filename
		     method close() = ()
		   end
		 ) in
	       fs # write_file [ `Create; `Truncate; `Binary ] path local;
	       Password.commit pwcache;
	       Password.save_entries pwcache;
	       `Ok ()
	    )
	with
	  | Ftp_client.FTP_timeout _ ->
	      `Timeout
	  | error ->
	      `Error(string_of_exn error)

      method api args =
        `Error "No API support for FTP servers"
		
    end
  )


let file_handler base_path : wtool_handler =
  let fs_lz = (* may result in an exception *)
    lazy(
      Netfs.local_fs
	~root:base_path
	()
    ) in
  ( object(self)
      method base_url = 
	Neturl.string_of_url
	  (Neturl.file_url_of_local_path base_path)

      method list path =
	try
	  let fs = Lazy.force fs_lz in
	  let files = fs # readdir [] path in
	  `Ok(List.filter
		(fun n -> n <> "" && n.[0] <> '.')
		files
	     )
	with
	  | e ->
	      `Error(string_of_exn e)

      method get tmpdir paths =
	let n = ref 0 in
	List.map
	  (fun path ->
	     try
	       let fs = Lazy.force fs_lz in
	       let local_filename =
		 Godi_file.addname
		   (Godi_file.tmpdir_path tmpdir)
		   ("wtool-" ^ string_of_int !n) in
	       Netchannels.with_out_obj_channel
		 (new Netchannels.output_channel
		    (Godi_file.open_out `Binary local_filename))
		 (fun out ->
		    Netchannels.with_in_obj_channel
		      (fs # read [ `Binary ] path)
		      (fun obj -> out # output_channel obj)
		 );
	       `Ok local_filename
	     with
	       | error ->
		   `Error(string_of_exn error)
	  )
	  paths

      method put local_filename path =
	try
	  let fs = Lazy.force fs_lz in
	  Netchannels.with_in_obj_channel
	    (new Netchannels.input_channel
	       (Godi_file.open_in `Binary local_filename))
	    (fun inp ->
	       Netchannels.with_out_obj_channel
		 (fs # write [ `Create; `Truncate; `Binary ] path)
		 (fun obj ->
		    obj # output_channel inp
		 )
	    );
	  `Ok ()
	with
	  | error ->
	      `Error(string_of_exn error)

      method api args =
        `Error "No API support for local file access"
    end
  )
  

let wtool_factory() =
  let timeout = ref 30.0 in
  let cuser = ref None in
  let store_passwords = ref false in
  let localbase = ref "/UNCONFIGURED" in
  ( object(self)
      method configure_timeout t = timeout := t
      method configure_user u = cuser := Some u
      method configure_store_passwords flag = store_passwords := flag
      method configure_localbase s = localbase := s
      method open_location base_url =
	if base_url <> "" && base_url.[0] = '/' then
	  (* Just interpret as local filename *)
	  file_handler base_url
	else
	  try
	    let u = 
	      Neturl.parse_url ~accept_8bits:true 
		(Neturl.fixup_url_string base_url) in
	    let have_user_in_url =
	      Neturl.url_provides ~user:true u in
	    let u0 =
	      Neturl.remove_from_url
		~user:true ~user_param:true ~password:true
		u in
	    let bu =
	      Neturl.string_of_url u0 in  (* to get consistent behavior *)
	    let proto = 
	      Neturl.url_scheme u in
	    let user =
	      if have_user_in_url then (
		if Neturl.url_user ~encoded:true u = "GODI_USER" then
		  !cuser
		else
		  Some(Neturl.url_user u)
	      )
	      else None in
	    let en_store = !store_passwords in
	    match proto with
	      | "http" | "https" ->
		  http_handler
                    ~localbase:!localbase
                    ~timeout:!timeout ~user ~en_store ~proto bu
	      | "ftp" ->
		  if user = None && use_web_proxy bu then
		    http_handler
                      ~localbase:!localbase
		      ~timeout:!timeout ~user ~en_store ~proto:"ftp" bu
		  else
		    ftp_handler
		      ~timeout:!timeout ~user ~en_store u0
	      | "file" ->
		  let path = Neturl.local_path_of_file_url u0 in
		  file_handler path
	      | _ ->
		  failwith ("Unsupported URL scheme: " ^ base_url)
	  with
	    | Neturl.Malformed_URL ->
		failwith ("Bad URL: " ^ base_url)
	
      method url_help =
	[ "http://[user@]host/path     (including WebDAV support)";
	  "https://[user@]host/path    (including WebDAV support)";
	  "ftp://[user@]host/path";
	  "file:///path"
	]

      method have_auth = true
    end
  )

let () =
  Ssl.init()
